package contextG

import org.codehaus.groovy.runtime.metaclass.MetaMethodIndex
import org.codehaus.groovy.runtime.MetaClassHelper
import org.codehaus.groovy.runtime.wrappers.Wrapper
import org.codehaus.groovy.reflection.ReflectionCache
import org.codehaus.groovy.runtime.metaclass.ClosureMetaMethod

/**
 * A layer is a collection of additional or overridden method and properties of several classes.
 * When a layer is activated, these methods are invoked instead of the original implementations.
 * 
 * The activation of a layer is usually performed during runtime and thus the activated layers 
 * can be computed. 
 */
class Layer {
    def name
    def overriddenProperties = [:]
     
    // a map (name -> layer) of all defined layers
    static layers = [:]
    // the current layers within a thread
    static activeLayers = [ initialValue: { [] } ] as ThreadLocal    
    
    static int anonymousLayerCounter = 0
    
    /**
     * Creates explicitly a layer with the given name.
     * In general, it is preferred to use define(Layer.<i>name</i>) instead
     */ 
    def Layer(name = null) {
        this.name = name ?: "Layer${anonymousLayerCounter++}"
        layers[name] = this
    }
    
    /**
     * Sets the specified property of the klass in the context of the layer.
     * If value is a Closure, it is treated as an overriding function with the given name.
     * 
     * This method should not be invoked directly and is automatically invoked in the DefinitionEnvironment.
     */
    def setProperty(klass, name, value) {        
        if(!(value instanceof Closure))
        	// plain property
            overriddenProperties.get(klass, [:])[name] = value
        else {
        	// method property
            def metaMethodIndex = getMethodIndex(klass)

            // creates a MetaMethod of the closure (basically just gives a name to the closure)
            def method = new ClosureMetaMethod(name, value)

            // now add it to the meta method index
            def header = metaMethodIndex.getHeader(klass)
            MetaMethodIndex.Entry e = metaMethodIndex.getOrPutMethods(name, header)
            if (method.isStatic()) 
                e.staticMethods = metaMethodIndex.addMethodToList(e.staticMethods, method)
            e.methods = metaMethodIndex.addMethodToList(e.methods, method)
        }
    }
        
    /**
     * Looks up a method for the given klass and name which can handle the given arguments.
     * Note that findMethod uses instances of classes as arguments and pickMethods the classes themselves.
     */
    def findMethod(klass, name, arguments, recursively = false) {
    	// look up for redefined method in method index
    	// methods can be added through setProperty in DefinitionEnvironment
        def methodIndex = getMethodIndex(klass)
        def e = methodIndex.getMethods(klass, name)
        
        // no entry yet
        if(!e) {
        	// looks for inherited redefinitions and save results (even missing methods)
        	if(recursively && klass.superclass) {
                def header = methodIndex.getHeader(klass)
                e = methodIndex.getOrPutMethods(name, header)

                def classes = MetaClassHelper.convertToTypeArray(arguments)
                def cacheEntry = new MetaMethodIndex.CacheEntry(params: classes, 
                    method: findMethod(klass.superclass, name, arguments, true))
                e.cachedMethod = cacheEntry
                return cacheEntry.method
        	}
        	return null
        }
    	
        // no methods for this klass and name
        if(!e.methods) return null

    	// look for a direct match in the cache
        def cacheEntry = e.cachedMethod
        if (cacheEntry && sameClasses(cacheEntry.params, arguments, e.methods instanceof MetaMethod))
            return cacheEntry.method

        // now continue with a fuzzy match and cache any results
        def classes = MetaClassHelper.convertToTypeArray(arguments)
        cacheEntry = new MetaMethodIndex.CacheEntry(params: classes, 
            method: chooseMethod(name, e.methods, classes))
        e.cachedMethod = cacheEntry
        cacheEntry.method
    }
    
    /**
     * Looks up a method for the given klass and name which can handle the given argument classes.
     * Note that findMethod uses instances of classes as arguments and pickMethods the classes themselves.
     */
    def pickMethod(klass, name, argumentClasses) {
        def methodIndex = getMethodIndex(klass)
        def e = methodIndex.getMethods(klass, name)
        // no methods for this klass and name        
        if(!e || !e.methods) return null

    	// look for a direct match in the cache
        def cacheEntry = e.cachedMethod
        if (cacheEntry && cacheEntry.params == argumentClasses)
            return cacheEntry.method

        // now continue with a fuzzy match and cache any results
        cacheEntry = new MetaMethodIndex.CacheEntry(params: argumentClasses, 
            method: chooseMethod(name, e.methods, argumentClasses))
        e.cachedMethod = cacheEntry
        cacheEntry.method        
    }

    /**
     * Returns the value of the layered property or null if none exists
     */
    def getProperty(klass, name) {
        overriddenProperties.get(klass)?.get(name)
    }

    /**
     * Returns true if there is a layered property or null if none exists.
     */
    def hasProperty(klass, name) {
        overriddenProperties.get(klass)?.containsKey(name)
    }
    
    String toString() {
        "Layer $name"
    }
    
    /*
     * Convenience method for multiple activation: with(l1 + l2) { ... }
     */
    def plus(Layer layer) {
    	[this, layer]
    }

    /**
     * Returns the next layer which overrides the method of the given class.
     * The signature of the method is specified by the name and the types of the arguments.
     * If afterLayer is null, the first activated layer is returned. Otherwise every layer preceding afterLayer and afterLayer itself are skipped.
     */
    static getOverridingLayer(klass, name, arguments, afterLayer) {
        def layers = activeLayers.get()
        if(afterLayer)
            layers = layers.subList(layers.indexOf(afterLayer) + 1, layers.size())
        layers.find { it.findMethod(klass, name, arguments) }
    }

    /**
     * Returns the next layer which overrides the method of the given class.
     * The signature of the method is specified by the name and the argument classes.
     * If afterLayer is null, the first activated layer is returned. Otherwise every layer preceding afterLayer and afterLayer itself are skipped.
     */
    static getOverridingLayer(klass, name, Class[] argumentClasses, afterLayer) {
        def layers = activeLayers.get()
        if(afterLayer)
        	layers = layers.subList(layers.indexOf(afterLayer) + 1, layers.size())
        layers.find { it.pickMethod(klass, name, argumentClasses) }
    }

    static NO_SUCH_PROPERTY = new Object()
    /**
     * Returns the property value of the first layer overriding the property of the given class.
     */
    static getOverridingProperty(klass, propertyName) {
        for(layer in activeLayers.get()) 
            if(layer.hasProperty(klass, propertyName)) 
                return layer.getProperty(klass, propertyName)
        NO_SUCH_PROPERTY
    }

    /**
     * Sets the property value of the first layer overriding the property of the given class.
     * If no layer has declared the property, this method returns false.
     */
    static setOverridingProperty(klass, propertyName, value) {
        for(layer in activeLayers.get()) 
            if(layer.hasProperty(klass, propertyName)) {
                layer.setProperty(klass, propertyName, value)
                return true
            }
        false
    }
    
    /**
     * Executes the closure and stores every assignment as overridden methods or properties.
     */
    def defineLayer(Closure closure) {
    	defineLayer(this, closure)
    }

    /**
     * Activates this layer and executes the closure.
     */
    def with(Closure closure) {
        with(this, closure)
    }

    /**
     * Deactivates this layer and executes the closure.
     */
    def without(Closure closure) {
        without(this, closure)
    }

    /**
     * Executes the closure and stores every assignment as overridden methods or properties.
     */
    static defineLayer(layer, closure = null) {
        if(!layer) throw new IllegalArgumentException("no layer specified")
        // replace stub and register it
        if(layer instanceof LayerStub) 
            layers[layer.name] = layer = new Layer(layer.name)
        if(closure)
            new DefinitionEnvironment(layer: layer, declaringClass: closure.delegate).applyDefinition(closure)
    }
    
    /**
     * Activates the layer or the list of layers and executes the closure.
     */
    static with(List<Layer> additionalLayers, Closure closure) {      
    	// unify list, delete duplicates
        if(additionalLayers instanceof List)
            additionalLayers = additionalLayers.flatten().unique()
    	
        // check if it contains any layer
        if (additionalLayers.size() == 0) 
            return closure.call() 
        
        // trivial case: layers are already activated in the given order
        def currentActiveLayers = activeLayers.get()
        if (additionalLayers.size() < currentActiveLayers.size() && additionalLayers == currentActiveLayers[0..additionalLayers.size()])  
            return closure.call() 
        
        // check if there is any stub (non-defined Layer)
        for(layer in additionalLayers)      
            if(layer instanceof LayerStub)
                throw new IllegalArgumentException("Layer ${layer.name} is used before definition")
        
        // save old list
        def clonedActiveLayers = currentActiveLayers.clone()
        // insert the layers or move them to the front
        currentActiveLayers.removeAll(additionalLayers)
        // add layers in reversed order indicating that later elements preceed previous elements
        currentActiveLayers.addAll(0, additionalLayers.reverse())
        
    	// execute closure and restore old settings, even if any exception occurred
        try {
            return closure.call() 
        } finally {
            activeLayers.set(clonedActiveLayers);
        }
    }

    
    /**
     * Deactivates the layer or the list of layers and executes the closure.
     */
    static without(List<Layer> excludedLayers, closure) {
    	// unify list
        if(excludedLayers instanceof List)
            excludedLayers = excludedLayers.flatten()
        else excludedLayers = [excludedLayers]

        // check if it contains any layer
        if (excludedLayers.size() == 0) 
            return closure.call() 
            
        def currentActiveLayers = activeLayers.get();
        if(excludedLayers.size() == 1 && !currentActiveLayers.contains(excludedLayers[0])) 
            return closure.call() 
            
        // check if there is any stub (non-defined Layer)
        for(layer in excludedLayers)      
            if(layer instanceof LayerStub)
                throw new IllegalArgumentException("Layer ${layer.name} is used before definition")

        // save old list and remove the layers
        def clonedActiveLayers = currentActiveLayers.clone()
        currentActiveLayers.removeAll(excludedLayers)
        
    	// execute closure and restore old settings, even if any exception occurred
        try {
            return closure.call() 
        } finally {
            activeLayers.set(clonedActiveLayers);
        }
    }
    
    static initialize(object) {
        addInstanceMethods(object)
    }
        
    static {
    	// global method lookup should be avoided because of performance issues
    	// may be needed when classes are initialized before the Layer class
    	// ExpandoMetaClass.enableGlobally()
        Layer.metaClass.'static'.propertyMissing << { 
            if(!layers[it])
                layers[it] = new LayerStub(name: it)
            layers[it]
        }
        addInstanceMethods()
        
        // add static methods
        Class.metaClass.layer << Layer.&defineLayer
        ['with', 'without'].each { methodName ->
        	// list versions
        	Class.metaClass."$methodName" << Layer.&"$methodName"
        	// with 1 up to 10 expanded Layer parameters
        	def metaMethod = Layer.metaClass.respondsTo(null, methodName).find { it.parameterTypes.size() > 1 }
        	1.upto(10) { layerCount ->
	            Class.metaClass."$methodName" << new VariableLayerMethodClosure(Layer, metaMethod, layerCount) 
        	}
        }
    }
    
    static pseudoProperties = [new LayerMetaProperty('with', Layer.&with), 
                               new LayerMetaProperty('without', Layer.&without), 
                               new LayerMetaProperty('layer', Layer.&defineLayer)]
    static addInstanceMethods(objects = Object) {
        objects.each { object ->
            def meta = object.metaClass
            if(!meta.respondsTo(object, "without")) {
                if(meta instanceof ExpandoMetaClass) {
                    meta.layer << Layer.&defineLayer


                    ['with', 'without'].each { methodName ->
                    	// list versions
                    	meta."$methodName" << Layer.&"$methodName"
                    	// with 1 up to 10 expanded Layer parameters
                    	def metaMethod = Layer.metaClass.respondsTo(null, methodName).find { it.parameterTypes.size() > 1 }
                    	1.upto(10) { layerCount ->
                    	meta."$methodName" << new VariableLayerMethodClosure(object, metaMethod, layerCount) 
                    	}
                    }
                } else {
                    try {
                        pseudoProperties.each { meta.addMetaBeanProperty(it) }
                    } catch(Exception e) {}
                }
            }
        }
    }
    
    //
    // Multimethod resolutions extracted from MetaClassImpl
    // Basically, the next methods chooses the method which fits the given parameters the best
    //

    private getMethodIndex(klass) {
        def methodIndex = overriddenProperties.get(klass, [:])['$methods']
        if(!methodIndex)
            overriddenProperties.get(klass)['$methods'] = methodIndex = 
                new MetaMethodIndex(ReflectionCache.getCachedClass(klass))
        methodIndex
    }
    
    private sameClasses(params, arguments, weakNullCheck) {
        if (params.length != arguments.length)
            return false        

        for (int i = params.length-1; i >= 0; i--) {
            Object arg = arguments[i];
            if (arg) {
                if (arg instanceof Wrapper) {
                    if (params[i] != arg.type)
                        return false
                }
                else if (params[i] != arg.class)
                    return false
            } else if (!weakNullCheck)
                return false
        }

        return true;
    }
    
    private chooseMethod(methodName, methodOrList, arguments) {
        if (methodOrList instanceof MetaMethod) 
            return MetaClassHelper.isValidMethod(methodOrList, arguments) ? methodOrList : null

        def methods = methodOrList
        int methodCount = methods.size()
        if (methodCount <= 0) return null
        if (methodCount == 1) 
            return MetaClassHelper.isValidMethod(methods[0], arguments) ? methodOrList : null
                
        if (arguments == null || arguments.length == 0) 
            return MetaClassHelper.chooseEmptyMethodParams(methods)        
        if (arguments.length == 1 && arguments[0] == null)
            return MetaClassHelper.chooseMostGeneralMethodWith1NullParam(methods)
                
        def matchingMethods = methods.toList().findAll { 
            MetaClassHelper.isValidMethod(it, arguments) 
        }
        
        if (!matchingMethods) return null
        if (matchingMethods.size() == 1) 
            return matchingMethods.get(0)
        
        chooseMostSpecificParams(methodName, matchingMethods, arguments)
    }
    
    private chooseMostSpecificParams(name, matchingMethods, arguments) {
        def matchesDistance = Long.MAX_VALUE, matches = []
        
        matchingMethods.each { method ->
            def paramTypes = MetaClassHelper.getParameterTypes(method)
            def dist = MetaClassHelper.calculateParameterDistance(arguments, paramTypes)
            if (dist == 0) return method
            
            if(dist < matchesDistance) matches.clear()
            if(dist <= matchesDistance) {
                matches.add(method)
                matchesDistance = Math.min(dist, matchesDistance)
            }            
        }
        
        if(!matches) return null
        if (matches.size() == 1) return matches[0]

        //more than one matching method found --> ambigous!
        String msg = "Ambiguous method overloading for method "
        msg += theClass.getName() + "#" + name
        msg += ".\nCannot resolve which method to invoke for "
        msg += InvokerHelper.toString(arguments)
        msg += " due to overlapping prototypes between:"
        for (Iterator iter = matches.iterator(); iter.hasNext();) {
            Class[] types = MetaClassHelper.getParameterTypes(iter.next()).getNativeParameterTypes()
            msg += "\n\t" + InvokerHelper.toString(types)
        }
        throw new GroovyRuntimeException(msg);
    }
}

/*
 * Used to simulate variable parameter lists for with and without.
 * Because the last parameter is always a closure, the built-in constructs 'Layer... layers' cannot be used.
 * Instead, several overloaded versions are generated through the VariableLayerMethodClosure class.
 * Each VariableLayerMethodClosure accepts layerCount+1 parameters: layerCount layers and 1 closure.
 */
class VariableLayerMethodClosure extends Closure {
	def method

	public VariableLayerMethodClosure(Object owner, MetaMethod method, int layerCount) {
		super(owner);
		this.method = method;
		this.@parameterTypes = [Layer] * layerCount + Closure;
		this.@maximumNumberOfParameters = this.@parameterTypes.size();
	}

	public Object doCall(Object... arguments) {
		// extract the layers and invoke with/without with a list of layers and the closure
		return method.invoke(null, arguments[0..-2], arguments[-1]);
	}
}

/*
 * Placeholder for a undefined layer; generated through Layer."LayerName".
 */
class LayerStub {
    String name
}

/**
 * LayerMetaProperty is used to add the when, with, and without methods to MetaClasses 
 * which are not ExpandoMetaClasses.
 * Most MetaClasses support adding MetaBeanProperties to them during runtime and treat Closures as Methods.
 * However, this involves quite an overhead.
 */
class LayerMetaProperty extends MetaBeanProperty {
    def closure
    
    LayerMetaProperty(name, closure) {
        super(name, Object, new ClosureMetaMethod(name, closure), null)
        this.closure = closure
    }
    
    def getProperty(object) {
        closure
    }
}