package contextG
import groovy.util.GroovyTestCase

/*
 * Compilation of the different usages to confirm that ContextG passes all basic tests.
 * More detailed descriptions are given in the respective examples.
 */
class LayerTest extends GroovyTestCase {
    void setUp() {
        Layer.initialize(this) 
    }
    
    void testBasicLayerActivation() {
        Employer HPI = new Employer(name: "HPI", address: "14482 Potsdam")
        assertEquals "Name: HPI", HPI.toString()
        with (Layer.Address) {
            assertEquals "Name: HPI; Address: 14482 Potsdam", HPI.toString()
        }
        layer (Layer.Employment)
        with (Layer.Employment) {
            assertEquals "Name: HPI", HPI.toString()
        }
    }
    
    void testNestedLayers() {
        Employer HPI = new Employer(name: "HPI", address: "14482 Potsdam")
        // without layer
        assertEquals "Name: HPI", HPI.toString()
        // now add address
        with (Layer.Address) {
        	assertEquals "Name: HPI; Address: 14482 Potsdam", HPI.toString()
        }
        // no redefinition in this layer for Employer class
        with (Layer.Employment) {
        	assertEquals "Name: HPI", HPI.toString()
        }

        Person somePerson = new Person(name: "Arvid Heise", address: "12345 Berlin", employer: HPI)
        // address and employment layer add additional information
        assertEquals "Name: Arvid Heise", somePerson.toString()
        with (Layer.Address) {
            assertEquals "Name: Arvid Heise; Address: 12345 Berlin", somePerson.toString()
        }
        with (Layer.Employment) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI", somePerson.toString()
        }
        
        // depending of activation order this results in different order of added information
        with (Layer.Address) {
            with (Layer.Employment) {
            	// call stack would be:
        		// 1. [Layer.Employment] Person.toString(), invokes proceed()
        		// 2. [Layer.Address] Person.toString(), invokes proceed()
                // 3. Person.toString() returns "Name: Arvid Heise"
                // 2. [Layer.Address] Person.toString(), appends address and returns
                // 1. [Layer.Employment] Person.toString(), appends employer and returns
            	assertEquals "Name: Arvid Heise; Address: 12345 Berlin; [Employer] Name: HPI", somePerson.toString()
            }
        }
        
        with (Layer.Employment) {
            with (Layer.Address) {
            	// layer methods are invoked in reverse order
            	assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
            }
        }
        // shorter form for previous nesting
        with ([Layer.Employment, Layer.Address]) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
        }
        with (Layer.Employment + Layer.Address) {
            assertEquals "Name: Arvid Heise; [Employer] Name: HPI; Address: 12345 Berlin", somePerson.toString()
        }
    }    

    void testLayerDeactivation() {
        Employer HPI = new Employer(name: "HPI", address: "14482 Potsdam")
        Person somePerson = new Person(name: "Arvid Heise", address: "12345 Berlin", employer: HPI)
        with (Layer.Address) {
            without (Layer.Address) {
                assertEquals "Name: HPI", HPI.toString()
            }
        }
        with ([Layer.Address, Layer.Employment]) {
            without (Layer.Employment) {
                assertEquals "Name: Arvid Heise; Address: 12345 Berlin", somePerson.toString()
            }
        }
    }

    void testRecursion() {
        def m = new FibMath()
        assertEquals 89, m.fib(10)
    	with(FibMath.CountCalls) {
            assertEquals 177, m.fib(10)
        }
    }
    
    void testMultiMethods() {
		def m = new MultiMethod()
		// shows unlayered multi method resolution with different types of first paramters
		
		// direct hits
		// first param is int, chooses method with int param
		assertEquals "method(int param1, param2)", m.method(1, 1)
		// same with double...
		assertEquals "method(double param1, param2)", m.method(1d, 1)
		// and same with string...
		assertEquals "method(String param1, param2)", m.method("1", 1)
		
		// a long cannot be coerced to an int but a double is a good alternative representation
		assertEquals "method(double param1, param2)", m.method(1L, 1)
		// cannot use a specialized method
		assertEquals "method(param1, param2)", m.method([], 1)
		
		with(Layer.OverloadSecondParam) {
			// even though there are appropriate specialized methods in the unlayered class...
			// ...only the layered method is chosen because layered methods take precedence
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1d, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method("1", 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1L, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method([], 1)
			
			// more difficult cases
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1, 1)
			assertEquals "layered: method = { param1, double param2 -> }", m.method(1, 1d)
			// no matching layered method was found and no coercion possible
			assertEquals "method(int param1, param2)", m.method(1, "1")
			// long can be coerced to double and therefore the second layered method can be applied
			assertEquals "layered: method = { param1, double param2 -> }", m.method(1, 1L)
			assertEquals "method(int param1, param2)", m.method(1, [])
		}
    }
    
    void testProperties() {
        def m = new FibMath()
    	with (FibMath.CountCalls) {
    	    m.reset()
            assertEquals 177, m.fib(10)
    	    m.calls = 0
            assertEquals 177, m.fib(10)
            assertEquals 354, m.fib(10)
        }
    }
    
    void testExternalDefinitions() {
        // TODO: not implemented yet
    }
    
    void testInheritance() {
		def b = new Base(), d = new Derived()

		// a layer overrides the instance methods of the same class
		assertEquals "base a", b.a()
		with(Layer.BaseLayer) {
			assertEquals "base layer a", b.a()
			assertEquals "base layer c", b.c()
		}
		
		// a layer does NOT override instance methods of derived classes if they are redefined there
		// this ensures that derived classes can always rely on their inner assumptions
		assertEquals "derived a", d.a()
		assertEquals "base c", d.c()
		with(Layer.BaseLayer) {
			// a is overriden in derived class
			assertEquals "derived a", d.a()
			// but c is not overriden and therefore layered
			assertEquals "base layer c", d.c()
		}
		// but of course methods can be redefined in their own layer
		with(Layer.DerivedLayer) {
			assertEquals "derived layer a", d.a()
			assertEquals "base c", d.c()
		}
		
		// same with static methods
		assertEquals "base b", Base.b()
		with(Layer.BaseLayer) {
			assertEquals "base layer b", Base.b()
		}
		
		// they are only redefined if the layers are defined after the normally executed method has been defined
		assertEquals "derived b", Derived.b()
		with(Layer.BaseLayer) {
			assertEquals "derived b", Derived.b()
		}
		with(Layer.DerivedLayer) {
			assertEquals "derived layer b", Derived.b()
		}
    }
}