package contextG

import static org.junit.Assert.*

/*
 * This test case demonstrates one of the groovy specifics: multi methods.
 * Because Groovy allows optional static typing, method overloading is possible.
 * In traditional static method overloading, the appropriate method is chosen at compile time.
 * For example in Java:
 * 
 * void overloadedMethod(String param) { System.out.println("param is string"); }
 * void overloadedMethod(Object param) { System.out.println("param is object"); }
 * void main() { overloadedMethod((Object) "testString"); }
 * Would print "param is OBJECT". Because at compile time, "testString" is casted to object.
 * 
 * In Groovy this same code works differently:
 * void overloadedMethod(String param) { System.out.println("param is string"); }
 * void overloadedMethod(Object param) { System.out.println("param is object"); }
 * void main() { overloadedMethod((Object) "testString"); }
 * Results in "param is string", even when the string is casted to an object.
 * 
 * This is result of the dynamic method lookup at runtime which also chooses the most appropriate method
 * for the given parameters.
 * 
 * The following demo shows exactly what the user can expect when layering multi methods.
 */
class MultiMethod {
    def method(param1, param2) {
        "method(param1, param2)"
    }

    def method(int param1, param2) {
        "method(int param1, param2)"
    }

    def method(double param1, param2) {
        "method(double param1, param2)"
    }
    
    def method(String param1, param2) {
        "method(String param1, param2)"
    }
    
    static {
        layer(Layer.OverloadSecondParam) {
            method = { param1, int param2 ->
                "layered: method = { param1, int param2 -> }"                
            } 

            method = { param1, double param2 ->
                "layered: method = { param1, double param2 -> }"                
            }
        }
    }
    
	static void main(args) {
		def m = new MultiMethod()
		// shows unlayered multi method resolution with different types of first paramters
		
		// direct hits
		// first param is int, chooses method with int param
		assertEquals "method(int param1, param2)", m.method(1, 1)
		// same with double...
		assertEquals "method(double param1, param2)", m.method(1d, 1)
		// and same with string...
		assertEquals "method(String param1, param2)", m.method("1", 1)
		
		// a long cannot be coerced to an int but a double is a good alternative representation
		assertEquals "method(double param1, param2)", m.method(1L, 1)
		// cannot use a specialized method
		assertEquals "method(param1, param2)", m.method([], 1)
		
		with(Layer.OverloadSecondParam) {
			// even though there are appropriate specialized methods in the unlayered class...
			// ...only the layered method is chosen because layered methods take precedence
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1d, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method("1", 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1L, 1)
			assertEquals "layered: method = { param1, int param2 -> }", m.method([], 1)
			
			// more difficult cases
			assertEquals "layered: method = { param1, int param2 -> }", m.method(1, 1)
			assertEquals "layered: method = { param1, double param2 -> }", m.method(1, 1d)
			// no matching layered method was found and no coercion possible
			assertEquals "method(int param1, param2)", m.method(1, "1")
			// long can be coerced to double and therefore the second layered method can be applied
			assertEquals "layered: method = { param1, double param2 -> }", m.method(1, 1L)
			assertEquals "method(int param1, param2)", m.method(1, [])
		}
	}
}